import os
import json
import time
import codecs
import pickle
import hashlib
import logging
from logging.config import dictConfig

from langchain import LLMChain
from langchain.llms import OpenAI
from langchain.chat_models import ChatOpenAI
from langchain.chat_models import AzureChatOpenAI
from langchain.embeddings.openai import OpenAIEmbeddings
from langchain.utilities import SerpAPIWrapper
from langchain.schema import ChatMessage
from langchain.agents import Tool
from langchain.agents import AgentType
from langchain.agents import initialize_agent
from langchain.docstore.document import Document
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain.vectorstores import Chroma
from langchain.vectorstores import FAISS
from langchain.document_loaders import Docx2txtLoader
from langchain.document_loaders import PyPDFLoader
from langchain.document_loaders import PyMuPDFLoader
from langchain.document_loaders import PDFMinerLoader
# from langchain.document_loaders import PyPDFium2Loader
from langchain.indexes.vectorstore import VectorstoreIndexCreator
from langchain.document_loaders import UnstructuredHTMLLoader
from langchain.document_loaders import DirectoryLoader

import chromadb
from chromadb.config import Settings

import fitz

from llmsdk.agents import agent_events
from llmsdk.lib.statemgmt import RedisMixin

try:
    from llmsdk.services.log import *
    logging.config.dictConfig(log_config)
except:
    logging.basicConfig()

class BaseLLMAgent(RedisMixin):
    """
    Class to define the base Hasper LLM agent
    """

    def __init__(self,
                 name,
                 cred,
                 agent_type,
                 platform="openai",
                 searchapi="serpapi",
                 statestore="redis",
                 memory_size=1000,
                 topk=7):
        """
        init the LLM query agent
        name: name of the agent
        cred: credentials object
        platform: name of the LLM platform backend to use
                default to OpenAI GPT platform for now, Azure is also supported
                will be extended in the future to suuport other models
        memory_size: how many tokens of memory to use when chatting with the LLM
        """

        # logging
        self.logger = get_logger()

        # name
        start_time = time.time()
        self.agent_name = name
        self.agent_type = agent_type
        self.agent_id = self._create_id(f"{self.agent_name}_{start_time}")

        # How many documents/chunks to respond to?
        self.topk = topk

        # statekey of the agent
        # the statekey is a pointer to the current state of the agent
        # this state can be persisted to some store
        # when needed, any previous persisted state
        # can be recalled and reassigned to the agent
        # for now, statekey is static
        # in the future, we can store and retrieve possibly multiple states
        self.statekey = f"sk:{self.agent_type}#{self.agent_name}"

        # creds
        self.cred = cred

        # LLM params
        self.platform = platform

        # init the llm and embeddings objects
        self.llm, self.embeddings = self._get_llm_objs(platform=self.platform,
                                                       cred=self.cred)


    ## helper functions

    def _create_id(self, string):
        """
        create a unique identifier from a string
        """
        return hashlib.md5(string.encode('utf-8')).hexdigest()

    def _log_event(self, event_name, duration=0, params={}):
        """
        format an event object to store in the agent's metadata
        event_name: name of the event to log
        duration: how long did this event take, only needed for events
                   that require some waiting (e.g. RTT to LLM APIs)
        params: any other info needed to be logged
        """

        ts = time.time() - duration # to get the actual start time
        event_id = self._create_id(f'{event_name}-{ts}')

        event = {
            "id": event_id,
            "agent": self.agent_name,
            "agent_type": self.agent_type,
            "agent_id": self.agent_id,
            "timestamp": round(ts, 3),
            "duration": round(duration, 3),
            "name": event_name,
            "params": params,
        }
        self.metadata['events'].append(event)

        # log the event
        self.logger.debug(event_name,
                             extra={
                                 'source': self.agent_name,
                                 'data': json.dumps(event, indent=4)
                             })


        return event

    def _get_llm_objs(self, platform, cred):
        # get the api key from creds
        api_key = self._get_api_key(cred, platform)

        # init the LLM
        if platform == "openai":

            # get the llm object
            # gpt-3.5-turbo
            llm = ChatOpenAI(temperature=0,
                             model="gpt-3.5-turbo",
                             openai_api_key=api_key,
                             request_timeout=20)

            # get the embeddings object
            embeddings = OpenAIEmbeddings(openai_api_key=api_key)

        elif platform == "azure":

            api_type = 'azure'
            api_version = '2023-05-15' # this may change in the future

            # get the llm object
            # based on deployment
            api_key = os.getenv("AZURE_OPENAI_KEY")
            api_base = os.getenv("AZURE_OPENAI_ENDPOINT") # your endpoint should look like the following https://YOUR_RESOURCE_NAME.openai.azure.com/
            deployment_name = os.getenv("AZURE_OPENAI_DEPLOYMENT_NAME_GPT") # this will correspond to the custom name you chose for your deployment when you deployed a model
            llm = AzureChatOpenAI(openai_api_base=api_base,
                                  openai_api_version=api_version,
                                  deployment_name=deployment_name,
                                  openai_api_key=api_key,
                                  openai_api_type=api_type,
                                  temperature=0,
                                  request_timeout=20)

            # get the embeddings object
            os.environ["OPENAI_API_KEY"] = api_key
            os.environ["OPENAI_API_TYPE"] = api_type
            os.environ["OPENAI_API_BASE"] = api_base
            os.environ["OPENAI_API_VERSION"] = api_version
            deployment_name = os.getenv("AZURE_OPENAI_DEPLOYMENT_NAME_EMBEDDING") # this will correspond to the custom name you chose for your deployment when you deployed a model
            embeddings = OpenAIEmbeddings(deployment=deployment_name,
                                          model="text-embedding-ada-002",
                                          chunk_size=16)

        else:
            llm = None
            embeddings = None

        return llm, embeddings

    def _load_search_agent(self, cred, searchapi, llm):
        """
        setup the search agent
        this agent will be used to run searches against search engines
        useful to get realtime answers to queries
        """
        # get the api key from creds
        api_key = self._get_api_key(cred, searchapi)

        # setup the search tool
        if searchapi == "serpapi":
            searchengine = SerpAPIWrapper(serpapi_api_key=api_key)
        else:
            searchengine = None
        if searchengine == None:
            return None

        tools = [
            Tool(
                name="Search",
                func=searchengine.run,
                description="Useful for when you need to answer questions about current events. Do not try to click on links. Input must be a search query."
            )
        ]
        # create the agent
        agent_chain = initialize_agent(tools,
                                       llm,
                                       agent=AgentType.ZERO_SHOT_REACT_DESCRIPTION,
                                       verbose=False)
        return agent_chain, searchengine

    def _get_path_source(self, fpath):
        """
        infer the format from the content path name
        """
        if os.path.isdir(fpath):
            return "dir"
        else:
            return fpath.split('.')[-1]

    def _get_api_key(self, cred, key):
        """
        get the API key from the cred
        """
        api_key = None
        if isinstance(cred, str):
            api_key = cred
        if isinstance(cred, dict) and key in cred:
            api_key = cred[key]
        return api_key

    def _err_msg(self, t):
        msgs = {
            "field": "I'm having trouble understanding. Try another way of wording your query."
        }
        return msgs.get(t, "Something went wrong, try your query again.")


    ##
    ## interfaces
    ##

    def get_metadata(self):
        """
        return metadata collected by the agent
        """
        return self.metadata


    ## state management

    def get_statekey(self):
        """
        return the statekey which points to the
        latest state of the agent
        """
        return self.statekey

    def clear_agent_state(self):
        """
        clear the memory context of the agent
        this is useful when we want to agent to
        start off a new QnA session
        """
        start_time = time.time()

        # clear the memories
        self.memory.clear()
        self.kg.clear()

        # reset the context and KG vars
        self.latest_context = []
        self.current_kg = []

        # make note of the clear action
        duration = time.time() - start_time
        params = {}
        event = self._log_event(agent_events._EVNT_STATECLEAR, duration, params=params)

        return

    def get_agent_state(self):
        """
        get the state of the agent
        """
        # get the memories
        memory = self.memory
        kg = self.kg

        # get the context and KG vars
        latest_context = self.latest_context
        current_kg = self.current_kg

        # setup the state object
        state = {
            "memory": memory,
            "kg": kg,
            "latest_context": latest_context,
            "current_kg": current_kg,
        }

        return state

    def store_agent_state(self):
        """
        store the agent state in some external storage
        this is useful when we want to later retrieve
        the state and re-animate an agent with the retrieved state
        """
        start_time = time.time()

        # get the state
        state = self.get_agent_state()

        # pickle it for transport
        state_pkl = codecs.encode(pickle.dumps(state), "base64").decode()

        # store it
        success = False
        statekey = self.statekey
        if self.state_store == "redis":
            cred = self.cred.get(self.state_store, {})
            success = self.store_state_redis(statekey, state_pkl, cred)
        elif self.state_store == "db":
            pass
        elif self.state_store == "disk":
            pass
        else:
            raise Exception(f"Unknown state store for agent, cannot store the agent's state for {statekey}")

        # make note of the action
        if success:
            duration = time.time() - start_time
            params = {
                "statekey": statekey,
                "store": self.state_store
            }
            event = self._log_event(agent_events._EVNT_STATESTORE, duration, params=params)

        return success

    def set_agent_state(self, statekey):
        """
        get a previously stored state
        and re-animate the agent with it
        """
        start_time = time.time()
        success = False

        # retrieve the state from storage
        state_pkl = None
        if self.state_store == "redis":
            cred = self.cred.get(self.state_store, {})
            state_pkl = self.retrieve_state_redis(statekey, cred)
        elif self.state_store == "db":
            pass
        elif self.state_store == "disk":
            pass
        else:
            raise Exception(f"Unknown state store for agent, cannot retrieve state for {statekey}")

        state = None
        if state_pkl:
            # re-animate our agent with the retrieved state
            # unpickle the state pickle
            try:
                state = pickle.loads(codecs.decode(state_pkl.encode(), "base64"))
            except:
                self.logger.error("Cannot re-animate agent with retrieved state, setting clear state",
                                     extra={
                                         'source': self.agent_name,
                                         'data': f"State pickle: {state_pkl}"
                                     })
                # clear the agent state
                self.clear_agent_state()

        if state:
            # re-hydrate state vars
            if state.get('memory'):
                self.memory = state['memory']
            if state.get('kg'):
                self.kg = state['kg']
            if state.get('latest_context'):
                self.latest_context = state['latest_context']
            if state.get('current_kg'):
                self.current_kg = state['current_kg']
            success = True

        # note the action
        duration = time.time() - start_time
        params = {
            "statekey": statekey,
            "success": success,
            "store": self.state_store
        }
        event = self._log_event(agent_events._EVNT_STATERESTORE, duration, params=params)

        return


    ## data loading

    def chunk_data(self, data, _format):
        """
        create chunks from the data
        and add any needed metadata
        """

        def cleanup_metadata(data):
            # take in a list of data document objects
            # and clean up metadata
            curr_source = None
            for i in range(0, len(data)):

                source = data[i].metadata['source']
                content = data[i].page_content

                # add bboxes for pdf docs
                if _format == "pdf":
                    pageno = data[i].metadata['page']
                    page = self.docs[source].load_page(pageno)
                    instances = page.search_for(content)
                    bboxes = []
                    for inst in instances:
                        bboxes.append([inst.x0, inst.y0,
                                       inst.x1, inst.y1])
                    data[i].metadata['bboxes'] = json.dumps(bboxes)

                # add other metadata
                data[i].metadata['file'] = source.split('/')[-1]
                if curr_source != source:
                    curr_source = source
                    chunk = 1
                data[i].metadata['chunk'] = chunk
                if _format == "str":
                    data[i].metadata['id'] = self._create_id(f"{source}-{self._create_id(content)}-{chunk}")
                else:
                    data[i].metadata['id'] = self._create_id(f"{source}-{chunk}")
                chunk += 1

                ##
                ## add any other custom metadata here
                ##

            return data

        # chunk the data
        text_splitter = RecursiveCharacterTextSplitter(chunk_size=self.chunk_size,
                                                       chunk_overlap=self.chunk_overlap)
        chunks = text_splitter.split_documents(data)

        # add metadata
        chunks = cleanup_metadata(chunks)

        return chunks

    def load_data(self, content, source=None, metadata={}, params={}):
        """
        set the datasource loader, loads and cleans the data
        content: data content (path, text) depending on type of source
        source:
            'dir' points to a folder with one or more files to load
            'pdf' points to a single PDF file to load
            'str' contains a text string to load
            None will try to infer automatically
        params: extra params needed for specific loaders
                    glob: what glob filter to use if source=='dir'
                    pdfloader: what type of pdf loader module to use if source=='pdf'
        metadata: any custom metadata to pass when source=='str'
        """

        start_time = time.time()
        success = False

        if not source:
            source = self._get_path_source(content)

        if source == 'dir':
            glob = params.get("glob", "**/*.*")
            loader = DirectoryLoader(content, glob=glob, recursive=True)
            data = loader.load()

        elif source == 'pdf':
            pdfloader = params.get("pdfloader", "pymupdf")
            if pdfloader == "pymupdf":
                loader = PyMuPDFLoader(content)
                data = loader.load()
            elif pdfloader == "pypdf":
                loader = PyPDFLoader(content)
                data = loader.load_and_split()
            elif pdfloader == "pypdfium2":
                loader = PyPDFium2Loader(content)
                data = loader.load()
            elif pdfloader == "pdfminer":
                loader = PDFMinerLoader(content)
                data = loader.load()
            else:
                data = None

            # make note of this document
            # we'll need to access it later for metadata
            if data:
                self.docs[content] = fitz.open(content)

        elif source == 'docx':
            loader = Docx2txtLoader(content)
            data = loader.load()

        elif source == 'str':
            # special handling for string inputs
            metadata['source'] = source
            data = [Document(page_content=content, metadata=metadata)]

        else:
            data = None

        # chunk the data
        if data:
            data = self.chunk_data(data, source)
            success = True

        # log that the data loader is ready
        duration = time.time() - start_time
        params = {
            "success": success,
            "source": source,
            "content": content,
            "params": params,
            "metadata": metadata,
        }
        event = self._log_event(agent_events._EVNT_DATA, duration, params=params)

        return data


    ## index management

    def get_index(self):
        """
        return the index object
        """
        return self.index

    def create_add_index_chroma(self, data, persist_directory=None):
        """
        Init the chromadb index and populate it with a set of documents
        """
        # init the ChromaDB client
        self.vdb_client = chromadb.Client(Settings(
            chroma_db_impl="duckdb+parquet",
            persist_directory=persist_directory # Optional, defaults to .chromadb/ in the current directory
        ))

        # make sure we're starting with a fresh db and collection
        self.vdb_client.reset()
        index = self.vdb_client.get_or_create_collection(name=self.index_name,
                                                         embedding_function=self.embeddings.embed_documents)

        # populate the collection
        self._add_docset_chroma(index, data)

        # persist the index to disk
        if persist_directory:
            self.vdb_client.persist()

        return index

    def add_to_index_chroma(self, data):
        """
        add document(s) to a chromadb index
        """
        # first, delete all existing docs from the same sources
        # as what we are adding, we don't want duplicates
        self._delete_docset_chroma(self.index, data)

        # now, add the new docs
        self._add_docset_chroma(self.index, data)

        # persist the db to disk
        self.vdb_client.persist()

        return

    def add_to_index(self, data):
        """
        add document(s) to the agent's index
        """
        start_time = time.time()

        if self.index:
            if self.index_store == 'chroma':
                self.add_to_index_chroma(data)
                # force topk count to be at max count of index
                self.topk = min(self.topk, self.index.count())
            else:
                raise Exception(f"{self.index_store} does not support adding document")
        else:
            raise Exception("No available index, cannot add document")

        # log that the doc is added
        if self.index:
            duration = time.time() - start_time
            params = {
                "n_items": len(data),
            }
            event = self._log_event(agent_events._EVNT_INDEXADD, duration, params=params)

        return

    def create_add_index(self, data, store='chroma', persist_directory=None, index_name=None):
        """
        create an index from a data source
        data: list of langchain Document() objects
        store: type of vectorstore to use (chroma, faiss, ...)
        """

        start_time = time.time()

        # note what store we are using and the index name
        self.index_store = store
        self.index_name = self.agent_name if not index_name else index_name

        # create the index
        if store == 'faiss':
            self.index = FAISS.from_documents(data, self.embeddings)
        elif store == 'chroma':
            self.index = self.create_add_index_chroma(data, persist_directory=persist_directory)
            # force topk count to be at max count of index
            self.topk = min(self.topk, self.index.count())
        else:
            self.index = None
            self.index_store = None

        # log that the index is ready
        if self.index:
            duration = time.time() - start_time
            params = {
                "store": store,
                "persist_directory": persist_directory,
                "index_name": self.index_name,
                "n_items": len(data)
            }
            event = self._log_event(agent_events._EVNT_INDEXCREATE, duration, params=params)

        return

    def load_index(self, persist_directory, index_name, store='chroma'):
        """
        load an already persisted index from a directory
        persist_directory: location of persisted index
        store: type of vectorstore to use (chroma, ...)
                only supports chroma for now
        """
        start_time = time.time()

        # make note of the store type
        self.index_store = store

        # load the index
        if self.index_store == 'chroma':
            self.vdb_client = chromadb.Client(Settings(
                                    chroma_db_impl="duckdb+parquet",
                                    persist_directory=persist_directory
                                ))
            self.index_name = index_name
            self.index = self.vdb_client.get_collection(name=self.index_name,
                                                          embedding_function=self.embeddings.embed_documents)
            # force topk count to be at max count of index
            self.topk = min(self.topk, self.index.count())
        else:
            self.index = None

        # log that the index is ready
        duration = time.time() - start_time
        params = {
            "store": self.index_store,
            "persist_directory": persist_directory,
        }
        event = self._log_event(agent_events._EVNT_INDEXLOAD, duration, params=params)

        return

    def get_index_stats(self):
        """
        return some stats about the agent's index
        """
        stats = None
        if self.index:
            try:
                stats = {
                    "name": self.index_name,
                    "store": self.index_store,
                    "n_items": self.index.count()
                }
            except:
                raise Exception("Index does not support stats")
        return stats

    def search_chromadb(self, query, k=7, include_metadata=True):
        """
        run a search against the chromadb index for a list of queries
        """
        # perform query
        results = self.index.query(query_texts=[query],
                                    n_results=k,
                                    where=None,
                                    where_document=None)
        # construct result docset
        docs = []
        for i in range(0, len(results['documents'][0])):
            page_content = results['documents'][0][i]
            metadata = results['metadatas'][0][i]
            distance = results['distances'][0][i]
            metadata['distance'] = distance
            doc = Document(page_content=page_content, metadata=metadata)
            docs.append(doc)

        return docs

    def _add_docset_chroma(self, index, data):
        """
        populate the index
        """
        if len(data) == 0:
            # cannot add anything
            return

        ids = [doc.metadata.get("id", self._create_id(doc.page_content)) for doc in data]
        docs = [doc.page_content for doc in data]
        metas = [doc.metadata for doc in data]

        index.add(
            documents=docs,
            metadatas=metas,
            ids=ids)

        return

    def _delete_docset_chroma(self, index, data):
        """
        delete from the index where source matches incoming data
        """
        ids = list(set([doc.metadata.get('id') for doc in data]))
        if len(ids) == 1:
            where_clause = {"id": ids[0]}
        else:
            where_clause = {
                "$or": [{"id": _id} for _id in ids]
            }
        index.delete(
            where=where_clause
        )
        return


    ## document search

    def get_similar_docs(self, query, topk=7):
        """
        get top-K similar docs from an index given a query
        query: query string
        index: index object
        topk: number of top-K similar docs to matching query to return
        """

        if self.index:
            if self.index_store == 'faiss':
                docs = self.index.similarity_search(query,
                                                    k=topk,
                                                    include_metadata=True)
            elif self.index_store == 'chroma':
                docs = self.search_chromadb(query,
                                            k=topk,
                                            include_metadata=True)
            else:
                docs = None

        return docs



import re
import os
import sys
import py_compile
import ast
from setuptools import setup, find_packages

_version_re = re.compile(r"VERSION\s+=\s+(.*)")

thisdir = os.path.dirname(__file__)
readme = open(os.path.join(thisdir, "README.rst")).read()

with open("enrichsdk/__init__.py", "rb") as f:
    version = str(
        ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1))
    )

setup(
    name="enrichsdk",
    version="5.0.7",
    description="Enrich Developer Kit",
    long_description=readme,
    url="http://github.com/pingali/scribble-enrichsdk",
    author="Venkata Pingali",
    author_email="pingali@scribbledata.io",
    license="All rights reserved",
    scripts=[],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "ruptures",
        "wheel",
        "click>=7.1.2",
        "aioitertools==0.8.0",

        "typing-extensions==4.8.0",
        "glob2==0.7",
        "httpx",
        "h2",
        "chardet==4.0.0",
        "requests==2.26.0",
        "requests-oauthlib==0.8.0",
        "pytest>=5.0.7",
        "pandas>=1.3.5,<1.5",
        "distributed==2.30.1",
        "idna==2.8",
        "coverage==5.5",
        "flake8",
        "raven==6.6.0",
        "python-json-logger==2.0.4",
        "python-dateutil>=2.8.1",
        "pydantic<=1.10.10",

        # numpy/statsmodels issues
        "numpy<=1.23.0",

        # Celery breaks without these..
        # https://github.com/celery/celery/issues/7607
        "s3fs>=0.5.1,<=0.5.2",
        "fsspec>=0.8.5,<=0.8.7",

        "botocore>=1.19.51,<=1.24.21",
        "aiobotocore>=1.2.1,<=1.2.2",

        "colored==1.3.5",
        #"flask-multistatic==1.0",
        "humanize==0.5.1",
        "pytz>=2020.1",
        #"Flask==2.0.3",
        "Jinja2>=3.0.3",
        "pytest-cov",
        "Markdown>=2.9.10",
        "prompt-toolkit>3.0.1,<3.1.0",
        "pyarrow>=0.9.0",
        "cytoolz>=0.9.0.1",
        "jsonschema>=3.2.0",
        "scipy<=1.10.0",
        "seaborn",
        #"flask_cors",
        #'moto>=1.3.14',
        "packaging<=20.0",
        "dask>=2.17.0,<20.0.0",
        "prefect>=0.15.7,<0.15.11",
        "distro>=1.4.0",
        "gcsfs==0.6.0",
        "jupyter-core>=4.12.0",
        "nbformat>=5.1.2",
        "tzlocal>=2.0.0",
        "texttable",
        "pykafka",
        "redis",
        "gitpython",
        "logstash_formatter",
        "pyhive",
        "pyfiglet",
        "sqlalchemy>=1.4.0",
        "kafka-python==2.0.2",
        "pykafka==2.8.0",
        "papermill>=2.3.4",
        "sqllineage",
        "google-cloud-logging",
        "unidecode",
        "faker",
        "xlsxwriter",
        "cryptography",
        "pymongo",
        "great-expectations==0.15.39",
        "pandas-stubs==1.5.3.230321",
        "anyio<=3.5.0",
        "openai",
        "pretty-html-table",

        # prophet dependencies
        "plotly>=4.0.0",
        "prophet==1.1.3",

        # classification
        "imblearn==0.0",

        # Boto
        #"boto3==1.16.0"

        # Azure dependencies
        "msgraph-sdk==1.0.0a13",
        "msal>=1.24.0,<=1.25.0"

    ],
    entry_points={
        "console_scripts": [
            "enrichpkg=enrichsdk.scripts.enrichpkg:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



from ..agents.extractor import *

import json

def get_contract_profilespec():

    import json

    # example of how to write methods to modify
    # profilespec queries
    def mod_query_duration(query, params):
        """
        modify the duration query based on the value
        populated in the billing frequency column
        """
        freqs = {
            "daily": "days",
            "weekly": "weeks",
            "monthly": "months",
            "quarterly": "months",
            "semi-annually": "months",
            "annually": "years",
            "yearly": "years",
        }

        # get the inputs we need
        billing_freq = params.get("billing_freq")
        # check if we have them and can proceed
        if not billing_freq:
            return query

        # modify the duration query
        billing_freq = billing_freq.lower()
        freq = freqs.get(billing_freq)
        if freq:
            query = f"{query} in {freq}?"

        return query

    # example of how to write methods to post-process LLM response
    def pp_companies(answer):
        """
        post-process the answer for the
        queries on parties involved
        """
        try:
            answer = json.loads(answer)
        except:
            pass
        return answer

    query_spec = {
        "name": "acme",
        "query_set": [
            {
                "name": "industry",
                "query": "what industry vertical does the contract deal with? respond with only the type of industry",
                "query_alts": [
                    "what industry is mentioned? respond with only the type of industry"
                ],
                "use_alts": "on-fail",
                "postprocess": None,
                "fill_columns": ["industry_vertical"]
            },
            {
                "name": "companies",
                "query": "which two companies is this contract between? format your response as a json list",
                "postprocess": pp_companies,
                "fill_columns": ["company_1", "company_2"]
            },
            {
                "name": "contract type",
                "query": "what type of contract does the document represent? choose from one of the following options: service agreement, master services agreement, purchase order, change order, subscription, saas services agreement, ammendment",
                "postprocess": None,
                "fill_columns": ["contract_type"]
            },
            {
                "name": "start date",
                "query": "what is the start date of the contract? respond with only a date.",
                "query_alts": [
                    "what is the effective date of the contract? respond with only a date"
                ],
                "use_alts": "on-fail",
                "postprocess": None,
                "fill_columns": ["commencement_date"]
            },
            {
                "name": "end date",
                "query": "what is the end date of the contract? respond with only a date.",
                "postprocess": None,
                "fill_columns": ["termination_date"]
            },
            {
                "name": "billing frequency",
                "query": "what is the billing frequency mentioned in the contract? choose from one of the following options: daily, weekly, monthly, quarterly, semi-annually, annually, adhoc",
                "postprocess": None,
                "fill_columns": ["billing_freq"]
            },
            {
                "name": "duration",
                "query": "what is the duration of the contract",
                "query_mod": {
                    "method": mod_query_duration,
                    "inputs": ["billing_freq"],
                    "apply_to": "first"
                },
                "query_alts": [
                    "what is the duration of the contract?"
                ],
                "use_alts": "always",
                "postprocess": None,
                "fill_columns": ["duration"]
            },
            {
                "name": "billing currency",
                "query": "what is the billing currency for the contract? respond with only the currency code",
                "query_alts": [
                    "what cuurency is the cost mentioned in? respond only with the code"
                ],
                "use_alts": "on-fail",
                "postprocess": None,
                "fill_columns": ["billing_currency"]
            },
            {
                "name": "billable amount",
                "query": "what is the total billable value of the contract?",
                "query_alts": [
                    "what is the cost mentioned?"
                ],
                "use_alts": "on-fail",
                "postprocess": None,
                "fill_columns": ["billable_amount"]
            }
        ]
    }

    return query_spec

def contract_query(profilespec, filename, cred=None):

    name = profilespec['name']
    persist_directory = "chromadb123"

    # create an agent
    agent = LLMQuerierExtractor(name=f"{name}_contract_agent", cred=cred)

    # point it to a document to read
    if filename.endswith("pdf"):
        agent.read_document(source="pdf",
                            content=filename,
                            persist_directory=persist_directory)
    elif filename.endswith("txt"):
        agent.read_document(source="str",
                            content=open(filename).read(),
                            metadata={
                                "source": filename,
                            },
                            persist_directory=persist_directory)
    else:
        raise Exception("Unsupported input file. Only pdf/txt are supported")

    # run it against the profilespec
    result = agent.process_spec(profilespec)

    return result, agent.get_metadata()



"""
File Operations
^^^^^^^^^^^^^^^

"""
import os
import sys
import abc
import shutil
import numpy as np
import pandas as pd
from enrichsdk import Trigger
from datetime import datetime
import logging

logger = logging.getLogger("app")


class FileOperationsBase(Trigger):
    """
    Base class for a FileOperations transform. For now only
    one action is supported 'copy'. More actions will be
    added in future.

    Example::

            {
                "transform": "FileOperations",
                "enable": true,
                "dependencies": {
                   ....
                },
                "args": {
                    "actions": [
                        {
                            "action": "copy",
                            "src": "%(output)s/%(runid)s/profile.sqlite",
                            "dst": "%(data_root)s/shared/campaigns/profile_daily/profile.sqlite",
                            "backupsuffix": ".backup"
                        },
                     ]
                }
            }

    """

    @classmethod
    def instantiable(cls):
        return False

    def __init__(self, *args, **kwargs):
        super(FileOperationsBase, self).__init__(*args, **kwargs)
        self.name = "FileOperationsBase"
        self.outputs = {}
        self.dependencies = {}

    def preload_clean_args(self, args):
        """
        Clean when the spec is loaded...
        """

        # Update the args
        args = super().preload_clean_args(args)

        # Sanity check...
        if not isinstance(args, dict):
            raise Exception("args should be a dictionary")

        if ("actions" not in args) or (not isinstance(args["actions"], list)):
            raise Exception("actions is missing or invalid")

        for a in args["actions"]:
            if not isinstance(a, dict):
                raise Exception("Each action spec should be a dictionary")
            supported = ["copy"]
            if a["action"] not in supported:
                raise Exception("Unsupported action")

            if a["action"] == "copy":
                if ("src" not in a) or ("dst" not in a):
                    raise Exception(
                        "Each copy action spec should specify a src and dst"
                    )

        return args

    def process(self, state):
        """
        Run the computation and update the state
        """
        logger.debug(
            "{} - process".format(self.name),
            extra=self.config.get_extra({"transform": self.name}),
        )

        msg = ""
        actions = self.args["actions"]
        for a in actions:
            action = a["action"]

            # Pass any global variables...
            srcbase = self.config.get_file(a["src"], extra=self.args)
            dstbase = self.config.get_file(a["dst"], extra=self.args)

            if "files" not in a:
                copyactions = [{"src": srcbase, "dst": dstbase}]
            else:
                copyactions = []
                for f in a["files"]:
                    copyactions.append(
                        {
                            "src": os.path.join(srcbase, f),
                            "dst": os.path.join(dstbase, f),
                        }
                    )

            backupsuffix = a.get("backupsuffix", ".backup")
            data_root = self.config.get_file("%(enrich_data_dir)s")
            for ca in copyactions:
                src = ca["src"]
                dst = ca["dst"]

                if not os.path.exists(src):
                    raise Exception("Could not find source: {}".format(src))

                if os.path.exists(dst):
                    backupsuffix = datetime.now().strftime(backupsuffix)
                    backupdst = dst + backupsuffix
                    if os.path.exists(backupdst):
                        if os.path.isdir(backupdst):
                            shutil.rmtree(backupdst)
                        else:
                            os.remove(backupdst)
                    os.rename(dst, dst + backupsuffix)

                try:
                    os.makedirs(os.path.dirname(dst))
                except:
                    pass

                # Handle the directory names
                if os.path.isdir(src):
                    shutil.copytree(src, dst)
                else:
                    shutil.copy(src, dst)

                msg += "Copy: {} => {}\n".format(
                    os.path.relpath(src, data_root), os.path.relpath(dst, data_root)
                )

        logger.debug(
            "{} - Completed".format(self.name),
            extra=self.config.get_extra({"transform": self.name, "data": msg}),
        )

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        """
        Check to make sure that the execution completed correctly
        """
        pass


provider = FileOperationsBase



from .agent_events import *

from .basellm import *
from .qna import *
from .sadl import *
from .docgen import *
from .datagpt import *
from .extractor import *
from .mltclassifier import *
from .querymapper import *



import os
import json
import numpy as np
import pandas as pd
import time
import random
import sqlite3
from enrichsdk import Compute, S3Mixin
from datetime import datetime, timedelta, date
import logging
import unidecode
from collections import defaultdict

from enrichsdk.contrib.lib.transforms import note
from enrichsdk.contrib.lib.assets import profilespec

from faker import Faker

logger = logging.getLogger("app")


class SyntheticDataGeneratorBase(Compute):
    """
    Generate synthetic data given a specification

    Features of transform baseclass include:
        * Flexible configuration
        * Highlevel specification of synthetic data in each column
            * instance: pre-defined faker-based instances
            * distribution: pre-defined from statistical distributions
            * custom: custom defined in base/derived class
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "SyntheticDataGeneratorBase"
        self.description = "Generate synthetic data from a specification"
        self.testdata = {
            "data_root": os.path.join(os.environ["ENRICH_TEST"], self.name),
            "statedir": os.path.join(os.environ["ENRICH_TEST"], self.name, "state"),
            "conf": {"args": {}},
            "data": {},
        }
        self.fakeObj = None # for synthetic data generation, to be inited later on

    @classmethod
    def instantiable(cls):
        return False

    def init_faker_object(self):
        fakeObj = Faker()
        names           = defaultdict(fakeObj.name)
        addresses       = defaultdict(fakeObj.address)
        countries       = defaultdict(fakeObj.bank_country)
        phonenumbers    = defaultdict(fakeObj.msisdn)
        emaildomains    = defaultdict(fakeObj.domain_name)

        obj = {
            "faker": fakeObj,
            "generators": {
                "person-name": names,
                "country_2char": countries,
                "location-address": addresses,
                "phone_number": phonenumbers,
                "email_domain": emaildomains,
            }
        }

        return obj

    def get_handlers(self, spec):
        """
        Define various callbacks that take a dataframe, spec
        and compute.
        """
        return {}

    def get_spec_datasets(self, profile):
        if (not isinstance(profile, dict)) or ("datasets" not in profile):
            raise Exception("datasets not defined in profile")
        return profile["datasets"]

    ###########################################
    # Synthetic Data Generation
    ###########################################
    def generate_dataset(self, dataset):
        msg = ""

        # check dataset spec
        if (not isinstance(dataset, dict)):
            raise Exception("Dataset should be a valid dict")
        for p in ["columns"]:
            if (p not in dataset):
                err = f"Invalid dataset: {p} missing"
                raise Exception(err)

        name        = dataset["name"]
        columns     = dataset["columns"]
        store       = dataset.get("store", "store_default")
        nrecords    = dataset.get("nrecords", 100)

        msg += f"Generating dataset: {name}" + "\n"
        if dataset.get("nrecords") is None:
            msg += f"Number of records not specified, using default" + "\n"
        msg += f"Generating {nrecords} records" + "\n"

        # init an empty dict for the new dataset
        _d = {}

        # for each column that we need to generate
        for column in columns:
            for f in ["name", "generator", "method"]:
                if f not in column:
                    err = f"Invalid dataset-->column: {column}-->{f} missing"
                    raise Exception(err)

            msg += f"Generating column: {column['name']}" + "\n"

            success, _data, l_msg = self.generate_single_column(column, nrecords)
            msg += l_msg

            # no luck with data generation, move on
            if not success:
                continue

            # we have the synthetic data, collect it
            _d[column['name']] = _data


        # combine all generated columns into a single DF
        data = pd.DataFrame.from_dict(_d, orient='index').transpose()

        msg += note(data, f"Synthetic dataset: {name}") + "\n"

        logger.debug(
            f"Generated dataset: {name}",
            extra={"transform": self.name, "data": msg}
        )

        return data

    def generate_single_column(self, column, nrecords):
        success = False

        name        = column['name']
        generator   = column['generator']

        if generator == "faker":
            method = column['method']
            method_handler = None

            fakeObj = self.fakeObj['faker']

            if hasattr(fakeObj, method):
                method_handler = getattr(fakeObj, method)

            if not callable(method_handler):
                msg = f"Column: <{name}> -- Generator <{generator}> has no callable method <{method}>, skipping generation of column" + "\n"
                return success, None, msg

            # we have a callable handler
            data = []
            for _ in range(nrecords):
                data.append(method_handler())
            success = True

        elif generator == "distribution":
            msg = f"Column: <{name}> -- Distribution method not supported yet, skipping generation of column" + "\n"
            return success, None, msg

        elif generator == "custom":
            msg = f"Column: <{name}> -- Custom method not supported yet, skipping generation of column" + "\n"
            return success, None, msg


        msg = f"Generated {len(data)} records for column: {name}" + "\n"

        return success, data, msg


    ###########################################
    # Data Anonymization
    ###########################################
    def anonymize_dataset(self, spec, data):
        '''
        Anonymize a dataset given a spec
        '''
        msg = ""
        name    = spec['name']
        config  = spec['config']

        # whether to anonymize all columns or a spec is defined
        columns_to_anon = "all" if "columns" not in config else "given"

        df_columns = data.columns
        columns = config.get('columns', {})

        # run through each column and try to anonymize
        anon_columns = []
        for col_name, col_obj in columns.items():
            include = col_obj.get("include", "yes")
            if include != "yes":
                continue
            params = {}
            if col_name not in df_columns:
                msg += f"Column: {col_name} not found, skipping" + "\n"
            else:
                data[col_name], l_msg = self.anonymize_single_column(col_name, col_obj, data, params)
                anon_columns.append(col_name)
                msg += l_msg

        # drop the other columns is required by spec
        action = config.get("nontransformed", "retain")
        if action == "drop":
            data = data[anon_columns]

        msg += note(data, "Anonymized dataset") + "\n"

        logger.debug(
            f"Spec: {name} dataset anonymized",
            extra={"transform": self.name, "data": msg}
        )

        return data

    def anonymize_single_column(self, col_name, col_obj, data, params):
        '''
        Takes a dataset and anonymizes the specified column
        '''
        msg = ""

        # get the faker object
        fakeObj = self.fakeObj

        # setup handlers for the various anonymization types
        generators = {}
        # first for the lookup generators
        for g, lookup in fakeObj['generators'].items():
            generators[g] = {
                "type": "lookup",
                "handler": lookup
            }
        # then for the custom generators
        generators["numeric"] = {
            "type": "custom",
            "handler": "anon_numeric"
        }
        generators["email"] = {
            "type": "custom",
            "handler": "anon_email"
        }

        anon_type = col_obj['anon_type']
        _d = []
        if anon_type in generators:
            gen_type = generators[anon_type]['type']
            gen_handler = generators[anon_type]['handler']
            if gen_type == "lookup":
                # we call the apply only on the specific column
                data = data[col_name].apply(lambda x: gen_handler[x])
            else:
                handler = getattr(self, gen_handler)
                # we call the apply to the full dataframe, we may need other columns
                # return is only the relevant column
                data = handler(data, col_name, col_obj)
            msg += f"Column: {col_name} anonymized" + "\n"
        else:
            data = np.nan
            msg += f"Column: {col_name} -- No <{anon_type}> generator found, defaulting to NaN" + "\n"

        return data, msg

    #######################################
    # Custom generators go here
    #######################################
    def anon_numeric(self, data, col_name, column):
        '''
        Method to fuzz numeric data. Various fuzzing methods
        can be defined here.
        Input is the full dataframe, output is the relavant column being fuzzed.
        '''
        msg = ""

        method      = column.get("method", "perturb")
        params      = column.get("params", {})

        val = data[col_name]

        if method == "perturb":
            range = params.get("range", 0.05)
            val += random.uniform(-range*val, range*val)
        else:
            msg = f"Column {column['name']} -- Unknown method to anon column, setting default NaNs" + "\n"
            val = np.nan

        return val

    def anon_email(self, data, col_name, column):
        '''
        Method to anonymize email data. Can generate emails to match or not match
        data in some name field. Also respects original email domain distribution if required.
        Input is the full dataframe, output is the relavant column being anonymized.
        '''
        msg = ""

        match_names = column.get("match_names", True)
        if match_names is True:
            if "name_field" not in column:
                msg += f"Column {col_name} -- Unknown name field to match emails, setting random emails" + "\n"
                match_names = False
                return np.nan
            else:
                if column["name_field"] not in data.columns:
                    msg += f"Column {column['name']} -- name field not in dataframe, setting random emails" + "\n"
                    match_names = False
                    return np.nan

        def generate_email(fakeObj, row, col_name, column, match_names):
            # whitelist of email domains
            # if the origninal email is in this list, don't replace it
            # useful to maintain data distribution
            domain_whitelist = ['gmail.com',
                                'yahoo.com',
                                'hotmail.com',
                                'aol.com']

            email_col_name  = col_name
            orig_domain     = row[email_col_name].split('@')[1]

            # set the email domain first
            if column.get("dist", "yes") == "yes":
                # we need to ensure that the distribution of generated email domains
                # match what was present in the input
                # popular free email domains will carry over, while others will be
                # replaced with random domains while still retaining distribution
                if any([d==orig_domain for d in domain_whitelist]):
                    # retain the origninal domain name
                    domain = orig_domain
                else:
                    # get a new domain name
                    domain = fakeObj['generators']['email_domain'][orig_domain]
            else:
                # no need to match distribution of generated email domains
                domain = fakeObj['faker'].ascii_email().split('@')[1]

            if match_names is True:
                # we want to match the anon email with the name field
                name = row[column['name_field']]
                names = unidecode.unidecode(name).lower().split(' ')
            else:
                # we don't care about matching the anon email with the name field
                names = fakeObj['faker'].name().split(' ')

            firstname = names[0]
            lastname = names[-1]

            # possible variations of email
            nameparts = {
                1: f"{firstname}",
                2: f"{lastname}",
                3: f"{firstname}.{lastname}",
                4: f"{firstname}.{firstname[0]}.{lastname}",
                5: f"{firstname}.{lastname[0]}.{lastname}",
                6: f"{firstname}.{firstname[0]}.{lastname[0]}",
                7: f"{firstname}.{random.randint(1,10000)}",
                8: f"{firstname}_{random.randint(1,10000)}",
                9: f"{firstname}.{lastname}.{random.randint(1,10000)}",
            }
            choice = random.randint(1, len(nameparts))
            namepart = nameparts[choice]
            email = f"{namepart}@{domain}"

            return email

        val = data.apply(lambda x: generate_email(self.fakeObj, x, col_name, column, match_names), axis=1)

        return val




    ###########################################
    # Helper Functions
    ###########################################

    def update_frame(self, source, description, df, lineage=None):

        if isinstance(source, str):
            name = source
        else:
            name = source["name"]

        params = self.get_column_params(name, df)
        if lineage is not None:
            if isinstance(lineage, dict):
                params.append(lineage)
            else:
                params.extend(lineage)

        detail = {
            "name": name,
            "df": df,
            "frametype": "pandas",
            "description": description,
            "params": params,
            "transform": self.name,
            "history": [],
        }

        self.state.update_frame(name, detail)

    def read_s3_data(self, filename, params, **kwargs):
        # assume we have a resolved s3fs object
        s3 = self.args['s3']
        if s3.exists(filename):
            df = pd.read_csv(s3.open(filename), **params)
            return df
        return None

    def get_dataset_s3(self, spec):
        """
        Use the dataset object to read the dataset
        """
        run_date    = self.args['run_date']
        name        = spec["name"]
        config      = spec['config']


        for f in ["dataset", "filename"]:
            if f not in config:
                msg = f"{f} param needed in config " + "\n"
                logger.exception(
                    f"Dataset: {name} -- skipping", extra={"transform": self.name, "data": msg}
                )
                return None

        source      = config.get('source', 'registry')
        dataset     = config['dataset']

        dataset     = config["dataset"]
        pieces      = dataset.split('-')
        dataset_main = "-".join(pieces[:-1])
        dataset_subset = pieces[-1]
        filename    = config["filename"]
        params      = config.get("params", {})

        cache = self.args.get("cache", False)
        cachename = f"{dataset}-{run_date}"
        cachefile = f"cache/{self.name}-anonymizer-cache-" + cachename + ".csv"

        if cache:
            try:
                os.makedirs(os.path.dirname(cachefile))
            except:
                pass
            if os.path.exists(cachefile):
                msg = f"Location: {cachefile}" + "\n"
                df = pd.read_csv(cachefile, **params)
                msg += note(df, f"Cached {dataset}") + "\n"
                logger.debug(f"Read cached {name}", extra={"transform": self.name, "data": msg})
                return df

        if source == "registry":
            if not hasattr(self, "get_dataset"):
                raise Exception(
                    "get_dataset_s3 expects get_dataset method"
                )
            datasetobj = self.get_dataset(dataset_main) # this method should be defined in the derived class

            if hasattr(self, 'update_doodle'):
                self.update_doodle(datasetobj, filename)

            df, metadata = datasetobj.read_data(
                run_date,
                run_date,
                filename=filename,
                readfunc=self.read_s3_data,
                params=params,
            )
        elif source == "direct":
            params = {}
            df = self.read_s3_data(filename, params)
            metadata = { "files": [filename] }
        else:
            logger.exception(
                f"Dataset: {name} -- unknown source param: {source}, skipping", extra={"transform": self.name}
            )
            return None

        msg = note(df, f"Fresh {dataset}") + "\n"
        logger.debug(f"Read fresh {name}", extra={"transform": self.name, "data": msg})

        # Cache it for future use...
        if cache:
            df.to_csv(cachefile, index=False)

        # Insert lineage if possible
        lineage = None
        if ("files" in metadata) and (len(metadata["files"]) > 0):
            lineage = {
                "type": "lineage",
                "transform": self.name,
                "dependencies": [
                    {
                        "type": "file",
                        "nature": "input",
                        "objects": [metadata["files"][-1]],
                    },
                ],
            }

        if not self.state.has_frame(spec['name']):
            self.update_frame(spec, f"Dataset: {dataset}", df, lineage)

        return df


    def load_dataset(self, spec, datasets):

        name        = spec['name']
        generate    = spec.get('generate')
        config      = spec['config']
        dataset     = config['dataset']

        # first, get the dataset nature
        nature = None
        paths = datasets[dataset].paths
        for path in paths:
            if path['name'] == 'default':
                nature = path['nature']
                break

        try:
            if nature == "db":
                data = self.read_db_source(spec)
            elif nature == "s3":
                data = self.get_dataset_s3(spec)
            elif (
                (generate is not None)
                and (generate in handlers)
                and (callable(handlers[generate]))
            ):
                data = handlers[generate](spec)
            elif (generate is not None) and (hasattr(self, generate)):
                data = getattr(self, generate)(spec)
            else:
                raise Exception(f"Dataset: {name} -- Invalid specification")
        except:
            msg =  "Could not load data, either nature or handlers not valid" + "\n"
            logger.exception(
                f"Dataset: {name} -- Generation failed", extra={"transform": self.name, "data": msg}
            )
            raise

        return data

    def s3_store_result(self, dataset, data):

        name        = dataset['name']
        run_date    = self.args['run_date']
        s3          = self.args['s3']
        epoch       = time.time()

        # get the dataframe and
        # add additional columns
        df = data
        df["__run_date__"] = run_date

        # where are we storing it?
        file   = os.path.join(self.args['s3root'], f"{run_date}/{name}.csv")

        # write to s3
        with s3.open(file, 'w') as fd:
            df.to_csv(fd, index=False)

        msg = f"s3 location: {file}" + "\n"

        logger.debug(f"Wrote dataset: {name} to S3",
                        extra={"transform": self.name,
                                "data": msg})

    def store_result(self, dataset, data):
        name    = dataset['name']
        store   = dataset.get("store", "s3")

        if store == "s3":
            # Now store in s3
            self.s3_store_result(dataset, data)
        else:
            logger.exception(f"Unknown store for dataset: {name}",
                         extra={
                             'transform': self.name
                         })

        # Send a message to SQL export that it should export
        self.state.msgpush(from_transform=self.name,
                           to_transform="SQLExport",
                           data={
                               'name': 'syndata',
                               'frames': [dataset['name']]
                           })

        #################################
        # also write to SQLLite
        # remove this when we figure out
        # how to do it from the pipeline
        # problem is -- output dataframe names can be dynamic, based on spec
        #   run_date    = self.args['run_date']
        #   s3          = self.args['s3']
        #   r_sqlfile   = os.path.join(self.args['s3root'], f"{run_date}/syndata.sqlite")
        #   l_sqlfile   = os.path.join(self.args['s3root'], f"syndata.sqlite.latest")
        #   cachefile   = f"cache/syndata.sqlite"

        #   conn = sqlite3.connect(cachefile)
        #   data.to_sql(name, conn, if_exists='replace', index=False)

        #   s3.put(cachefile, r_sqlfile)
        #   s3.put(cachefile, l_sqlfile)

        #   logger.debug("Storing as SQLite",
        #                extra={
        #                    'transform': self.name,
        #                    'data': f"Location(s): {r_sqlfile}, {l_sqlfile}" + "\n"
        #                })

        #################################

    def process(self, state):
        """
        Run the computation and update the state
        """
        logger.debug(
            "Start execution", extra=self.config.get_extra({"transform": self.name})
        )

        # Will be used in other places..
        self.state = state

        # init the faker object for data generation
        self.fakeObj = self.init_faker_object()

        # Get the profile spec
        is_valid, profile, msg = profilespec.get_profile(self, "syntheticdata")
        if is_valid:
            logger.debug(
                f"Loaded profilespec",
                extra={"transform": self.name, "data": msg}
            )
        else:
            logger.error(
                f"Could not load profilespec",
                extra={"transform": self.name, "data": msg}
            )
            raise Exception("could not load profilespec")

        # get the dataset lookup table
        customer_datasets = profilespec.construct_dataset_list(self, profile)

        # Now go through each dataset and generate synthetic data for it
        for spec in specs:

            process_spec = True

            enabled = spec.get("enable", True)
            if not enabled:
                logger.debug(
                    f"Spec <{spec.get('name', 'NO NAME')}> not enabled, skipping.",
                    extra={"transform": self.name}
                )
                process_spec = False
                continue

            for f in ["name", "config"]:
                if f not in spec:
                    logger.error(
                        f"Spec has no {f} param set, skipping.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                    )
                    process_spec = False
                    break

            if process_spec == False:
                # something is wrong with this spec, skip it
                continue

            # process the spec
            name    = spec['name']
            action  = spec.get('action', 'anonymize')

            if action == 'generate':
                # we have a generate a synthetic dataset
                frametype = "synthetic"
                data = self.generate_dataset(spec)
            elif action == 'anonymize':
                # we have to anonymize a given dataset
                frametype = "anonymized"
                # frist, load it
                data = self.load_dataset(spec, customer_datasets)

                # then, anonymize it
                if data is not None:
                    data = self.anonymize_dataset(spec, data)
                else:
                    msg = "Could not anonymize dataset" + "\n"
                    logger.exception(
                        f"Spec: {spec['name']} -- skipping",
                        extra={"transform": self.name}
                    )
            else:
                logger.exception(
                    f"Unknown action param in spec, skipping spec: {spec['name']}",
                    extra={"transform": self.name}
                )

            # store the generated dataset
            if data is not None:
                self.store_result(spec, data)

                # update frame for pipline
                description = spec.get(f"desc -- {frametype}", f"{frametype.title()} generated dataset")
                lineage = {
                    "type": "lineage",
                    "transform": self.name,
                    "dependencies": [
                        {
                            "type": "file",
                            "nature": "input",
                            "objects": [spec.get("filename", "__NEW__")],
                        },
                    ],
                }
                self.update_frame(
                    spec,
                    description,
                    data,
                    lineage,
                )

        # Done
        logger.debug(
            "Complete execution", extra=self.config.get_extra({"transform": self.name})
        )

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        pass



"""

This module implements the DocGen microservice. Its purpose is to wrap
around the DocGen agent and provide an integration point from end applications.
The DocGen module enables users to generate long-form documents based on a set of input documents and a spec.

"""
import os
import sys
import re
import glob
import yaml
import json
import traceback
import time
import hashlib
import struct
import logging
import threading
from uuid import uuid4
from datetime import datetime
from fastapi import FastAPI, BackgroundTasks, Depends
from fastapi.concurrency import run_in_threadpool
from fastapi_utils.tasks import repeat_every

from llmsdk.agents.docgen import LLMDocGenerator
from llmsdk.lib import SafeEncoder

from .log import *
from .lib import *

# Override default uvicorn logger...
logging.config.dictConfig(log_config)
logger = get_logger()

supported_exts=['txt', 'htm', 'html', 'pdf', 'docx', 'doc']

app = FastAPI()

# Insert policy
policyspec = []
add_policy(app, policyspec)

async def get_docgen_agent(namespace, username, dataset,
                           fullpath, files,
                           metadata,
                           drop_index=False):

    return await get_generic_agent(LLMDocGenerator,
                                   dataset,
                                   fullpath,
                                   files,
                                   metadata,
                                   drop_index=False)

add_show_cache(app)
add_show_health(app)
add_check_index(app, "docgen")
add_build_index(app, "docgen", supported_exts)

async def generate_run(request_id):
    """
    Query the agent instance created

    Parameters
    ----------
    request_id: str
                UUID generated for the request

    Returns
    -------
    None
         Nothing is returned. Only the result cache is updated
    """

    cache = get_cache()

    if request_id not in cache:
        logger.error(f"Failure",
                     extra={
                         'request_id': "invalid",
                         "source": "service",
                     })
        cache[request_id] = {
            'status': "failure",
            "message": f"Invalid request id"
        }
        return

    # First get the params
    value = cache[request_id]

    try:

        params  = value['params']
        user    = params['user']
        dataset = params['dataset']
        profilespec = params['profilespec']
        context = params['context']
        namespace = params['namespace']
        exts = params['exts']

        stats['query_count'] += 1

        label = f"{user}_{dataset}"

        logger.debug(f"{label}: Getting agent",
                     extra={
                         "source": "service",
                         "user": user,
                         "dataset": dataset,
                         "request_id": request_id,
                         'data': json.dumps(value, indent=4, cls=SafeEncoder)
                     })

        details = await get_agent_details(namespace,
                                          user,
                                          dataset,
                                          exts=exts,
                                          get_task_specific_agent=get_docgen_agent)
        if details is None:
            query_update_result(request_id, {
                'status': "failure",
                "message": f"LLM Docgen Agent not found for {label}. See agent log",
                "result": {}
            })
            return

        agent    = details['agent']
        metadata = details['metadata']

        answer   = await run_in_threadpool(lambda: agent.generate_doc(profilespec))
        result = {
            'request_id': request_id,
            "answer": answer,
            "metadata": metadata
        }
        stats['query_success'] += 1
        stats['datasets'][label]['query_count'] += 1
        stats['datasets'][label]['query_success'] += 1
        query_update_result(request_id, {
            "status": "success",
            "result": result
        })
    except Exception as e:

        stats['query_failure'] += 1
        stats['datasets'][label]['query_count'] += 1
        stats['datasets'][label]['query_failure'] += 1
        logger.exception(f"Failed to run query",
                         extra={
                             "source": "service",
                             'request_id': request_id,
                             "user": params.get('user', "unknown"),
                             "dataset": params.get('dataset', "unknown"),
                         })

        query_update_result(request_id, {
            "status": "failure",
            "answer": f"Unable to construct the answer. Internal error. See the agent log",
            "result": {
                "answer": str(e)
            }
        })



@app.get("/generate/status")
async def generate_status(request_id: str):
    """
    Get the status of the query request

    Parameters
    ----------
    request_id: str
                UUID generated for the request

    Returns
    -------
    status: dict
         Dictionary with request id, status, message and data
    """
    # First get the params
    try:

        value: dict = {}

        cache = get_cache()

        if request_id not in cache:
            logger.error(f"Query Status: Invalid Request ID",
                         extra={
                             'request_id': request_id,
                             "source": "service",
                         })
            return {
                "request_id": request_id,
                "status": "failure",
                'answer': "Data not available. Either the request is invalid or the memory with that request has been cleared on restart"
            }

        value = cache[request_id]
        status = value.get('status', 'unknown')
        logger.debug(f"Query Status: {status}",
                     extra={
                         "source": "service",
                         'request_id': request_id,
                         "dataset": value.get('dataset', 'unknown'),
                         "user": value.get('user', 'unknown'),
                         'data': json.dumps(value,indent=4, cls=SafeEncoder)
                     })
        return {
            "request_id": request_id,
            "status": value['status'],
            'message': value.get('message', ""),
            'data': value.get('result',{})
        }
    except Exception as e:
        logger.exception("Query Status: failure",
                     extra={
                         "source": "service",
                         'request_id': request_id,
                         "user": value.get('user',"unknown"),
                         "dataset": value.get('dataset',"unknown"),
                     })
        return {
            "request_id": request_id,
            "status": "failure",
            "message": str(e),
            "data": {}
        }

@app.post("/generate")
async def generate(user: str,
                   dataset: str,
                   profilespec: dict,
                   background_tasks: BackgroundTasks,
                   context: str = "",
                   namespace="docgen",
                   policy: dict = {},
                   extra: dict = {}
):
    """
    Initiate a new request

    Parameters
    ----------
    user: str
          User (an unintepreted string)
    profilespec: dict
           Dictionary specifying the document to the generated
    namespace: str
           A namespace associated with this instance of the service. Multiple
    context: str
           Any additional text
    policy: dict
           List of policies to enforce
    extra: dict
           Pass any other information

    Returns
    -------
    status: dict
         Dictionary with request id, query, status, and answer

    """
    cache = get_cache()

    request_id = str(uuid4())

    label = f"{user}_{dataset}"
    taskdetails = get_buildtask_details(label)
    if taskdetails is None:
        logger.error(f"Index not built yet",
                     extra={
                         'request_id': request_id,
                         "source": "service",
                         "user": user,
                         "dataset": dataset
                     })
        return {
            "request_id": request_id,
            "status": "failure",
            "answer": "Index has not been constructed yet. Please trigger a build action"
        }

    params = {
        "user": user,
        "dataset": dataset,
        "context": context,
        "namespace": namespace,
        "exts": supported_exts,
        "profilespec": profilespec,
        "policy": policy
    }
    logger.debug(f"Building index for {label}",
                extra={
                    "source": "service",
                    "user": user,
                    "dataset": dataset,
                    "data": json.dumps(params, indent=4, cls=SafeEncoder)
                })

    exists = await check_agent_details(user, dataset)
    if not exists:
        logger.error(f"Index not been built yet",
                     extra={
                         "source": "service",
                         "user": user,
                         "dataset": dataset,
                     })
        return {
            "request_id": request_id,
            "status": "failure",
            "answer": "Index has not been constructed yet. Please trigger a build action or wait if pending"
        }

    cache[request_id] = {
        "status": "pending",
        "user": user,
        "dataset": dataset,
        "params": params
    }

    # Run the background task...
    background_tasks.add_task(generate_run, request_id)

    return {
        "request_id": request_id,
        "status": "pending",
        "data": {}
    }

@app.on_event("startup")
def app_startup():

    t = BuildAgentThread(get_docgen_agent)
    t.start()
    initialize_stats()



"""
https://docs.google.com/document/d/1ObnJqcwzVmcTt8gT12DE8lil8J2LZBI-/edit
"""
#######################################################################
# Known Issues (https://developer.microsoft.com/en-us/graph/known-issues/?search=)
# Getting all items by navigating drive is partially supported, use list instead
#######################################################################
import os
import sys
import json
import logging
import string
import urllib
import traceback
import time
from pathlib import Path
from urllib import request
from urllib.parse import unquote
from collections import defaultdict
from datetime import datetime

import requests

try:
    from ..utils import SafeEncoder
except:
    from enrichsdk.utils import SafeEncoder

logger = logging.getLogger("app")

__all__ = ['AzureSharepointClient']

class AzureSharepointClient():
    """
    Simple azure client that uses GraphQL interface
    """

    def __init__(self, cred):
        self.cred = cred

    def get_access_token(self):

        if hasattr(self, 'token'):
            return self.token

        cred = self.cred
        tenant_id = cred['tenant_id']
        client_id = cred['client_id']
        client_secret = cred['client_secret']

        #Get access token
        token_url = f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token"
        token_data = {
            "grant_type": "client_credentials",
            "client_id": client_id,
            "client_secret": client_secret,
            "scope": "https://graph.microsoft.com/.default"
        }

        try:
            token_response = requests.post(token_url, data=token_data)
            access_token = token_response.json()["access_token"]
        except:
            token_data['client_secret'] = token_data['client_secret'][:10]+"..."
            raise Exception("Unable to get token",
                            extra={

                            })

        return access_token

    def list_items(self, folder):

        access_token = self.get_access_token()

        site_id = folder['site_id']
        dirname = folder['name']

        # => List files in the library
        url = f"https://graph.microsoft.com/v1.0/sites/{site_id}/lists/{dirname}/items"
        headers = {
            "Authorization": f"Bearer {access_token}"
        }


        # => Walk through the pages to get the items
        all_items = []
        toprocess = [url]

        while len(toprocess) > 0:
            url = toprocess.pop(0)
            graph_result = requests.get(url=url, headers=headers)
            if graph_result and graph_result.text:
                json_resp = json.loads(graph_result.text)

                # There is a 'next page' of results, then include
                # them..
                if '@odata.nextLink' in json_resp:
                    page_url = json_resp['@odata.nextLink']
                    toprocess.append(page_url)

                if 'value' in json_resp:
                    all_items.extend(json_resp['value'])

        return all_items

    def write_item_metadata(self, item, local_path):

        metadatafile = local_path + ".metadata.json"
        with open(metadatafile, 'w') as fd:
            fd.write(json.dumps({
                "details": item,
                "size": os.path.getsize(local_path),
                "timestamp": datetime.now().replace(microsecond=0)
            }, indent=4, cls=SafeEncoder))

    def read_item_metadata(self, item, local_path):

        metadatafile = local_path + ".metadata.json"
        try:
            return json.load(open(metadatafile))
        except:
            pass
        return None

    def should_download(self, item, local_path):

        if not os.path.exists(local_path):
            return True

        metadata = self.read_item_metadata(item, local_path)
        if metadata is None:
            return True

        size          = metadata['size']
        last_modified = metadata['details']['lastModifiedDateTime']
        etag          = metadata['details']['eTag']

        condns = [(size == os.path.getsize(local_path)),
                   (last_modified == item['lastModifiedDateTime']),
                   (etag == item['eTag'])]
        return not all(condns)

    def download_drive_item(self, folder, item, item_id, dir_path, local_path):
        """
        Download one item
        """

        access_token = self.get_access_token()
        headers = {
            "Authorization": f"Bearer {access_token}"
        }

        site_id = folder['site_id']

        if dir_path is None:
            raise Exception("Invalid drive path")

        download_url = f'https://graph.microsoft.com/v1.0/sites/{site_id}/drive/items/{item_id}/content'

        r = requests.get(download_url, allow_redirects=True, headers=headers)

        if r.status_code >= 400:
            logger.error("Failed to download",
                         extra={
                             "data": f"url: {download_url}\nResponse: {r}"
                         })
            raise Exception("Failed to download")

        with open(local_path, 'wb') as fd:
            fd.write(r.content)

        if not os.path.exists(local_path):
            raise Exception("Error while retrieving")

        self.write_item_metadata(item, local_path)

    def upload_item(self, folder, parent_item, local_path, filename=None):
        """
        Download one item
        """

        access_token = self.get_access_token()
        headers = {
            "Authorization": f"Bearer {access_token}",
            "Content-Type":"application/binary",
            'Accept': 'application/json',
        }

        site_id = folder['site_id']

        if ((local_path is None) or (not os.path.exists(local_path))):
            raise Exception("Invalid file path")

        if filename is None:
            filename = os.path.basename(local_path)

        data = open(local_path, 'rb').read()
        parent_item_id = self.get_etag_id(parent_item['eTag'])
        upload_url = f'https://graph.microsoft.com/v1.0/sites/{site_id}/drive/items/{parent_item_id}:/{filename}:/content'

        r = requests.put(upload_url, data=data, headers=headers)

        if r.status_code >= 400:
            reason = ""
            if r.status_code == [409, 410]:
                reason = "Locked"
            elif r.status_code in [401, 403]:
                reason = "Permission"
            logger.error(f"Failed to upload. {reason}",
                         extra={
                             "data": f"url: {upload_url}\nResponse: {r}"
                         })
            raise Exception("Failed to upload")

        #else:
        #    print(r)
        #    print(json.dumps(r.json(), indent=4))

        return r.json()

    def delete_item(self, folder, item):
        """
        Download one item
        """
        access_token = self.get_access_token()
        headers = {
            "Authorization": f"Bearer {access_token}",
        }

        site_id = folder['site_id']
        delete_url = f'https://graph.microsoft.com/v1.0/sites/{site_id}/drive/items/{item["id"]}'

        r = requests.delete(delete_url, headers=headers)

        if r.status_code >= 400:
            logger.error("Failed to delete",
                         extra={
                             "data": f"url: {upload_url}\nResponse: {r}"
                         })
            raise Exception("Failed to delete")

        #else:
        #    print(r)
        #    print(r.content)

    def get_etag_id(self, etag) -> str:
        # "eTag": "\"a6c0f692-b303-442c-ad12-1201fc4072b9,1\""
        # "eTag": "\"{386A8B52-95F8-47D5-9B1D-DA7DF6BE6C51},1\"",
        eid = None
        etag = etag.replace("\"", '')
        etag = etag.replace("{", '').replace("}","")
        ids = etag.split(",")
        if len(ids) > 0:
            eid = ids[0].lower()
        return eid

    def move_item(self, folder, item, parent_item, new_name=None):
        """
        Move the item to new parent item
        """

        access_token = self.get_access_token()
        headers = {
            "Authorization": f"Bearer {access_token}",
            'Content-type': 'application/json',
            'Accept': 'application/json',
        }

        site_id = folder['site_id']

        if item is None or parent_item is None:
            raise Exception("Invalid object or new parent folder")

        item_id = self.get_etag_id(item['eTag'])
        parent_item_id = self.get_etag_id(parent_item['eTag'])
        if new_name is None:
            new_name = item['webUrl'].split('/')[-1]

        data = {
            "parentReference": {
                "id": parent_item['parentReference']['id']
            },
            "name": "test-new-name.docx"
        }

        move_url = f'https://graph.microsoft.com/v1.0/sites/{site_id}/drive/items/{item_id}'
        r = requests.patch(move_url, data=json.dumps(data), headers=headers)

        if r.status_code >= 400:
            #print(r)
            #print(r.content)
            logger.error("Failed to move file",
                         extra={
                             "data": f"url: {move_url}\nResponse: {r}"
                         })
            raise Exception("Failed to move")

        #else:
        #    print(r)
        #    print(json.dumps(r.json(), indent=4))

    def copy_item(self, folder, item, parent_item):
        """
        Move the item to new parent item
        """

        access_token = self.get_access_token()
        headers = {
            "Authorization": f"Bearer {access_token}",
            'Content-type': 'application/json',
            'Accept': 'application/json',
        }

        site_id = folder['site_id']

        if item is None or parent_item is None:
            raise Exception("Invalid object or new parent folder")

        item_id = self.get_etag_id(item['eTag'])
        parent_item_id = self.get_etag_id(parent_item['eTag'])

        # print(item_id, "->", parent_item_id)
        data = {
            "parentReference": {
                "id": parent_item_id,
            },
            "name": "test-new-name.docx"
        }

        copy_url = f'https://graph.microsoft.com/v1.0/sites/{site_id}/drive/items/{item_id}/copy'
        r = requests.post(copy_url, data=json.dumps(data), headers=headers)

        if r.status_code >= 400:
            # print(r)
            # print(r.content)
            logger.error("Failed to move file",
                         extra={
                             "data": f"url: {copy_url}\nResponse: {r}"
                         })
            raise Exception("Failed to move")

        # else:
        #    print(r)
        #    print(json.dumps(r.json(), indent=4))


    def download_items(self, folder, items, metadata_handler=None):
        """
        Download the items into the local_root. Items should be the result of
        """

        local_root = folder['local_root']
        root_url = folder['root_url']
        stats = defaultdict(int)

        msg = ""
        downloaded = []
        for item in items:
            if 'contentType' not in item:
                stats['error_no_content_type'] += 1
                continue
            content_type = item['contentType']['name']
            weburl = item['webUrl']
            dir_path = unquote(weburl.replace(root_url, ''))
            if len(dir_path) == 0:
                stats['error_invalid_url'] += 1
                continue
            if dir_path.startswith("/"):
                dir_path = dir_path[1:]
            full_path = os.path.join(local_root, dir_path)
            if content_type == 'Folder':
                stats['folders'] += 1
                os.makedirs(full_path, exist_ok=True)
            else:
                stats['files'] += 1
                etag = item['eTag']
                if etag is None:
                    stats['error_file'] += 1
                    stats['error_file_no_etag'] += 1
                    continue
                eid = self.get_etag_id(etag)
                if eid is None:
                    stats['error_file'] += 1
                    stats['error_file_no_eid'] += 1
                    continue

                if self.should_download(item, full_path):
                    try:
                        self.download_drive_item(folder,
                                                 item,
                                                 eid,
                                                 dir_path,
                                                 full_path)
                        msg += f"[Download] {full_path}\n"
                        stats['file_downloaded'] += 1

                        # Dump intermittent
                        downloaded.append(dir_path)
                        if len(downloaded) > 20:
                            logger.debug(f"Downloaded {len(downloaded)} files",
                                         extra={
                                             'data': "\n".join(downloaded)
                                         })
                            downloaded = []

                    except:
                        msg += f"[Error] {full_path}\n"
                        stats['error_download_exception'] += 1
                        break
                else:
                    stats['file_skipped_exists'] += 1

            if len(downloaded) > 20:
                logger.debug(f"Remaining {len(downloaded)} files",
                             extra={
                                 'data': "\n".join(downloaded)
                             })
        return stats, msg


    def create_delta(self):
        headers = {
            'Authorization': self.access_token
        }

        sharepoint_site_id = self.params['sharepoint_site_id']
        delta_url = f'https://graph.microsoft.com/v1.0/sites/{sharepoint_site_id}/drive/root/delta'

        result = requests.get(url=delta_url, headers=headers)

        return result

    def get_latest_changes(self):
        headers = {
            'Authorization': self.access_token
        }

        sharepoint_site_id = self.params['sharepoint_site_id']
        delta_updates_url = f'https://graph.microsoft.com/v1.0/sites/{sharepoint_site_id}/drive/root/delta?token='

        result = requests.get(url=delta_updates_url, headers=headers)

        return result


if __name__ == "__main__":

    # Follow these instructions to get these attributes
    # https://docs.google.com/document/d/1ObnJqcwzVmcTt8gT12DE8lil8J2LZBI-/edit

    logging.basicConfig(level=logging.DEBUG)

    site_id = "aa4ff333-7526-4ad5-bf4b-a35b05007b24"
    site_name = 'testsharepointsite'
    folder = {
        "name": "Documents",
        "site_id": site_id,
        "root_url": f"https://scribbledatainc.sharepoint.com/sites/{site_name}/Shared%20Documents",
        "local_root": "downloads"
    }

    # Client details
    tenant_id = os.environ['TEST_TENANT_ID'] # UUID
    client_id = os.environ['TEST_CLIENT_ID'] # UUID
    client_secret = os.environ['TEST_CLIENT_SECRET']

    credentials = {
        'tenant_id': tenant_id,
        'client_id': client_id,
        'client_secret': client_secret
    }

    azureclient = AzureSharepointClient(credentials)

    # => List items in the folder...
    items = azureclient.list_items(folder)
    # print(json.dumps(items, indent=4))

    download = False
    if download:
        stats, msg = azureclient.download_items(folder, items)
        error_keys = [k for k in stats.keys() if k.startswith("error_")]
        logger.debug(f"Completed w/ {len(error_keys)} errors",
                     extra={
                         'data': json.dumps(stats, indent=4) + "\n" + msg
                     })

    root_url = folder['root_url']

    # Move one folder (Test/Avengers -> Processed)
    processed_item = avengers_item = None
    for item in items:
        if item['webUrl'] == f"{root_url}/Processed":
            processed_item = item
        elif item['webUrl'] == f"{root_url}/Test/SystemDescriptionDocument.docx": # some document
            avengers_item = item

    print("Avengers")
    print(json.dumps(avengers_item, indent=4))

    print("Processed")
    print(json.dumps(processed_item, indent=4))

    # Now upload a file and delete it..
    new_item = azureclient.upload_item(folder,
                                       processed_item,
                                       "SystemDescriptionDocument (1).docx") # some document

    print("Sleeping for 30 seconds")
    proceed = input("Y/n")
    if proceed in ['y','Y','']:
        azureclient.delete_item(folder, new_item)




import os
import json
import numpy as np
import pandas as pd
import time
import random
import sqlite3
import tempfile
import hashlib
import shutil
from enrichsdk import Compute, S3Mixin
from datetime import datetime, timedelta, date
import logging
import unidecode
from collections import defaultdict
from sqlalchemy import create_engine, text as satext

from enrichsdk.contrib.lib.transforms import note
from enrichsdk.contrib.lib.assets import profilespec
from enrichsdk.lib.integration import send_html_email

logger = logging.getLogger("app")


class DataObserverBase(Compute):
    """
    Monitor an input data source given a spec

    Features of transform baseclass include:
        * Flexible configuration
        * Highlevel specification of observability:
            * specified data source
            * custom defined testing conditions for observability
            * custom defined output of observability results
            * notification of observability results on success/failure
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "DataObserverBase"
        self.description = "Monitor an input data source given a spec"
        self.testdata = {
            "data_root": os.path.join(os.environ["ENRICH_TEST"], self.name),
            "statedir": os.path.join(os.environ["ENRICH_TEST"], self.name, "state"),
            "conf": {"args": {}},
            "data": {},
        }

    @classmethod
    def instantiable(cls):
        return False

    def get_handlers(self, spec):
        """
        Define various callbacks that take a dataframe, spec
        and compute.
        """
        return {}

    ###########################################
    # Check method for observability
    ###########################################
    def process_spec(self, spec, data):

        if data is None:
            msg = "No dataset loaded" + "\n"
            logger.exception(
                f"Spec: {spec['name']} -- skipping",
                extra={"transform": self.name}
            )
            return None

        name = spec['name']
        config = spec['config']
        checks = config['checks']

        results = []
        checks_done = []
        checked_at = datetime.now()

        run_id = f"{name}-{checked_at}"
        run_id = hashlib.md5(run_id.encode('utf-8')).hexdigest()

        # check if checks object is a list
        if not isinstance(checks, list):
            logger.exception(
                f"Invalid config param -- check", extra={"transform": self.name, "data": json.dumps(config, indent=4)}
            )
            return None

        # for each observability check
        for check in checks:
            do_check = True

            for f in ['name', 'method']:
                if f not in check:
                    logger.exception(
                        f"Invalid config param -- check", extra={"transform": self.name, "data": json.dumps(config, indent=4)}
                    )
                    do_check = False
                    break
            if do_check is False:
                continue

            if not hasattr(self, check['method']):
                logger.exception(
                    f"No method defined -- skipping check", extra={"transform": self.name, "data": json.dumps(check, indent=4)}
                )
                continue

            # we now have the callback
            callback = getattr(self, check['method'])
            result = callback(data)

            # add additional columns of interest
            result['run_id'] = run_id
            result['spec'] = name
            result['check'] = check['name']
            result['checked_at'] = checked_at
            result['status'] = 'fail' if result['status'] is False else 'pass'

            # add this check to the list of completed checks
            checks_done.append(check['name'])

            # add result to master list
            results.append(result)

        # convert the list into a DF
        results = pd.DataFrame(results)

        msg = f"Total {len(checks_done)} checks done: {checks_done}" + "\n"
        msg += note(results, "Check results")
        logger.debug(
            f"Processed checks: {spec['name']}",
            extra={"transform": self.name, "data": msg}
        )

        # if no checks were possible, return None so further pipelined steps are not done
        results = None if len(results) == 0 else results

        return results


    ###########################################
    # Helper Functions
    ###########################################

    def update_frame(self, source, description, df, lineage=None):
        if isinstance(source, str):
            name = source
        else:
            name = source["name"]

        params = self.get_column_params(name, df)
        if lineage is not None:
            if isinstance(lineage, dict):
                params.append(lineage)
            else:
                params.extend(lineage)

        detail = {
            "name": name,
            "df": df,
            "frametype": "pandas",
            "description": description,
            "params": params,
            "transform": self.name,
            "history": [],
        }

        self.state.update_frame(name, detail)

    ###########################################
    # I/O Functions
    ###########################################

    def read_s3_data(self, filename, params, **kwargs):
        # assume we have a resolved s3fs object
        s3 = self.args['s3']
        if s3.exists(filename):
            df = pd.read_csv(s3.open(filename), **params)
            return df
        return None

    def get_dataset_s3(self, spec):
        """
        Use the dataset object to read the dataset
        """
        run_date    = self.args['run_date']
        name        = spec["name"]
        config      = spec['config']
        source      = config['source']

        for f in ["dataset", "filename"]:
            if f not in source:
                msg = f"{f} param needed in config source" + "\n"
                logger.exception(
                    f"Dataset: {name} -- skipping", extra={"transform": self.name, "data": msg}
                )
                return None

        dataset_type    = source['type']
        dataset         = source['dataset']
        pieces          = dataset.split('-')
        dataset_main    = "-".join(pieces[:-1])
        dataset_subset  = pieces[-1]
        filename        = source["filename"]
        params          = source.get("params", {})

        cache = self.args.get("cache", False)
        cachename = f"{dataset}-{run_date}"
        cachefile = f"cache/{self.name}-anonymizer-cache-" + cachename + ".csv"

        if cache:
            try:
                os.makedirs(os.path.dirname(cachefile))
            except:
                pass
            if os.path.exists(cachefile):
                msg = f"Location: {cachefile}" + "\n"
                df = pd.read_csv(cachefile, **params)
                msg += note(df, f"Cached {dataset}") + "\n"
                logger.debug(f"Read cached {name}", extra={"transform": self.name, "data": msg})
                return df

        if dataset_type == "registry":
            if not hasattr(self, "get_dataset"):
                raise Exception(
                    "get_dataset_s3 expects get_dataset method"
                )
            datasetobj = self.get_dataset(dataset_main) # this method should be defined in the derived class

            if hasattr(self, 'update_doodle'):
                self.update_doodle(datasetobj, source['filename'])

            df, metadata = datasetobj.read_data(
                run_date,
                run_date,
                filename=filename,
                readfunc=self.read_s3_data,
                params=params,
            )
        elif dataset_type == "direct":
            df = self.read_s3_data(filename, params)
            metadata = { "files": [filename] }
        else:
            logger.exception(
                f"Unknown source param: {dataset_type}, skipping", extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
            )
            return None

        msg = note(df, f"Fresh {dataset}") + "\n"
        logger.debug(f"Read fresh {name}", extra={"transform": self.name, "data": msg})

        # Cache it for future use...
        if cache:
            df.to_csv(cachefile, index=False)

        # Insert lineage if possible
        lineage = None
        if ("files" in metadata) and (len(metadata["files"]) > 0):
            lineage = {
                "type": "lineage",
                "transform": self.name,
                "dependencies": [
                    {
                        "type": "file",
                        "nature": "input",
                        "objects": [metadata["files"][-1]],
                    },
                ],
            }

        if not self.state.has_frame(spec['name']):
            self.update_frame(spec, f"Dataset: {dataset}", df, lineage)

        return df


    def get_dataset_db(self, spec):
        name = spec['name']
        config = spec['config']
        source = config['source']

        testmode = self.args.get('testmode', False)
        env      = 'test' if testmode else 'prod'

        # Get the db engine
        engine_name = source['params'][env]['engine']
        engine = self.engines[engine_name]

        # Get the query
        query = source['params'][env]['query']
        if query.endswith('.sql'):
            # we have a file containing the query
            sqlfile = os.path.join(self.scriptdir, query)
            query = open(sqlfile).read()

        df = self.read_db_source(engine, query)

        # Now the input load...
        lineage = {
            "type": "lineage",
            "transform": self.name,
            "dependencies": [
                {
                    "type": "database",
                    "nature": "input",
                    "objects": [source['dataset']],
                },
            ],
        }

        self.update_frame(spec, f"Dataset: {name}", df, lineage)

        return df

    def read_db_source(self, engine, query):
        # Run the query
        df = pd.read_sql_query(satext(query), con=engine)

        return df


    def load_dataset(self, spec, datasets):

        name        = spec['name']
        generate    = spec.get('generate')
        config      = spec['config']
        source      = config['source']
        dataset     = source['dataset']
        dataset_type = source.get('type')

        if dataset_type == None:
            msg = f"Spec name: name" + "\n"
            msg += f"Config: {json.dumps(config, indent=4)}" + "\n"
            logger.exception(f"Unknown dataset type, skipping", extra={"transform": self.name, "data": msg})
            return None

        # get the dataset nature
        nature = source.get('nature')
        if dataset_type == "direct":
            if nature == None:
                msg = f"Spec name: name" + "\n"
                msg += f"Config: {json.dumps(config, indent=4)}" + "\n"
                logger.exception(f"Unknown dataset nature, skipping", extra={"transform": self.name, "data": msg})
                return None
        elif dataset_type == "registry":
            paths = datasets[dataset].paths
            for path in paths:
                if path['name'] == 'default':
                    nature = path['nature']
                    break
        elif dataset_type == "db":
            nature = "db"
        else:
            msg = f"Spec name: name" + "\n"
            msg += f"Config: {json.dumps(config, indent=4)}" + "\n"
            logger.exception(f"Unknown dataset type, skipping", extra={"transform": self.name, "data": msg})
            return None


        try:
            if nature == "db":
                data = self.get_dataset_db(spec)
            elif nature == "s3":
                data = self.get_dataset_s3(spec)
            elif (
                (generate is not None)
                and (generate in handlers)
                and (callable(handlers[generate]))
            ):
                data = handlers[generate](spec)
            elif (generate is not None) and (hasattr(self, generate)):
                data = getattr(self, generate)(spec)
            else:
                raise Exception(f"Dataset: {name} -- Invalid specification")
        except:
            msg =  "Could not load data, either nature or handlers not valid" + "\n"
            logger.exception(
                f"Dataset: {name} -- Generation failed", extra={"transform": self.name, "data": msg}
            )
            raise Exception(msg)

        msg = note(data, f"Dataset: {name}")
        logger.debug(
            f"Loaded dataset: {name}", extra={"transform": self.name, "data": msg}
        )

        return data

    def s3_store_result(self, spec, data):
        name        = spec['name']
        run_date    = self.args['run_date']
        s3          = self.args['s3']
        epoch       = time.time()

        # add additional columns
        data["__run_date__"] = run_date

        # where are we storing it?
        file = os.path.join(self.args['s3root'], f"{run_date}/{epoch}__{name}.csv")

        # write to s3
        with s3.open(file, 'w') as fd:
            data.to_csv(fd, index=False)

        msg = f"s3 location: {file}" + "\n"

        logger.debug(f"Wrote check results to S3",
                        extra={"transform": self.name,
                                "data": msg})

    def db_store_result(self, spec, results, data):
        name    = spec['name']
        config  = spec['config']
        store   = config['store']

        testmode = self.args.get('testmode', False)
        env      = 'test' if testmode else 'prod'

        # Get the db engine
        engine_name = store['params'][env]['engine']
        engine = self.engines[engine_name]

        # write results to db
        table_name = store['params'][env]['table']
        ret = results.to_sql(table_name,
                            engine,
                            if_exists='append',
                            index=False)
        # write data to db
        data_table_name = name.replace('-','_')
        ret = data.to_sql(data_table_name,
                            engine,
                            if_exists='append',
                            index=False)

        msg = f"DB engine: {engine}" + "\n"

        logger.debug(f"Wrote monitor results to DB",
                        extra={"transform": self.name,
                                "data": msg})

    def store_result(self, spec, results, data):
        name    = spec['name']
        config  = spec['config']
        store   = config['store']

        for f in ["sink", "params"]:
            if f not in store:
                logger.exception(
                    f"Store has no {f} param, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                )
                return

        run_id = results['run_id'].values[0]
        data['run_id'] = run_id

        sink = store['sink']
        if sink == "s3":
            # store in s3
            self.s3_store_result(spec, results, data)
        elif sink == "db":
            # store in db
            self.db_store_result(spec, results, data)
        else:
            logger.exception(f"Unknown store for dataset: {name}",
                         extra={
                             'transform': self.name
                         })

    ###########################################
    # Notification Functions
    ###########################################

    def check_notify_default(self, notify_set, notify_obj, spec, results):
        # default method to check if a notification needs to be sent
        name    = spec['name']
        config  = spec['config']
        store   = config['store']

        sink        = store['sink']
        channel     = notify_set['channel']
        notifier    = notify_set['name']
        send_on     = notify_set.get('send_on', 'fail')
        freq        = notify_set.get('freq', 'always')
        set_id      = notify_obj.get('set_id', '__ALL__')

        check_name      = results['check'].values[0]
        check_status    = results['status'].values[0]

        do_notify   = False

        # set time between notifications
        if freq == 'always':
            hours_between_notifications = 0
        else:
            try:
                hours_between_notifications = float(freq)
            except:
                # invalid notification frequency in spec
                msg = f"Spec: {name}" + "\n"
                msg += f"Target: {notify_set}" + "\n"
                msg += f"Defaulting to 1 hour" + "\n"
                logger.exception(f"Invalid notification frequency",
                             extra={
                                 'transform': self.name,
                                 'data': msg
                             })
                hours_between_notifications = 6

        # setup the send_on list
        send_on = ["pass", "fail"] if send_on=='both' else [send_on]

        msg = f"Channel: {channel}" + "\n"
        msg += f"Frequency: {freq}" + "\n"
        msg += f"Hours betwwen: {hours_between_notifications}" + "\n"
        msg += f"Send on: {send_on}" + "\n"
        logger.debug(f"Got notification params",
                     extra={
                         'transform': self.name,
                         'data': msg
                     })

        msg = ""
        if sink == "s3":
            # get result+notification data from s3
            pass
        elif sink == "db":
            # get result+notification data from db
            # get the environment we are in
            testmode = self.args.get('testmode', False)
            env      = 'test' if testmode else 'prod'
            # Get the db engine
            engine_name = store['params'][env]['engine']
            engine = self.engines[engine_name]
            table_name = store['params'][env]['table']
            # get the latest notification sent for this spec and notification channel
            query = f"""SELECT *
                        FROM {table_name}
                        WHERE `spec`='{name}'
                            AND `check`='{check_name}'
                            AND `set_id`='{set_id}'
                            AND `notifier`='{channel}'
                            AND `notified`=1
                        ORDER BY `notified_at` DESC
                        LIMIT 1
                        """
            msg += f"Engine: {engine}" + "\n"
            msg += f"Query: {query}" + "\n"
            try:
                df = self.read_db_source(engine, query)
            except:
                # empty dataframe
                msg += f"Failed reading DB, defaulting to empty DF" + "\n"
                df = pd.DataFrame()
        else:
            logger.exception(f"Unknown store for {name}",
                         extra={
                             'transform': self.name,
                             'data': json.dumps(store, indent=4)
                         })
            return False

        msg += note(df, "Last notification DF") + "\n"
        msg += f"check_status={check_status}, send_on={send_on}" + "\n"
        logger.debug(f"Last notification status",
                     extra={
                         'transform': self.name,
                         'data': msg
                     })

        if len(df) > 0:
            # find time between last notification and now
            last_notified_at = pd.Timestamp(df['notified_at'].values[0])
            time_elapsed = (datetime.now() - last_notified_at).total_seconds() / (60*60) # convert to hours
            # check if time elapsed since last notification is more than spec
            if time_elapsed > hours_between_notifications and check_status in send_on:
                do_notify = True
        else:
            if check_status in send_on:
                do_notify = True

        return do_notify

    def notify_via_slack(self, target, spec):
        # trivial notification method to simulate
        # notification by slack
        return True

    def notify_via_email(self, notify_obj, spec):
        # notification via email
        name        = spec['name']
        receivers   = notify_obj['receivers']
        bcc         = notify_obj.get('bcc', [])
        cc         = notify_obj.get('cc', [])
        reply_to    = notify_obj.get('reply_to', [])
        content     = notify_obj.get('content')
        subject     = notify_obj.get('subject')
        sender      = notify_obj.get('sender')
        smtpcred    = notify_obj.get('smtpcred', {})
        retries     = notify_obj.get('retries', 1)
        data        = notify_obj['data']
        send_attachment = notify_obj.get('send_attachment', False)
        notify_success  = False

        # create the content if not available
        if content == None:
            content = f"Ran ledger balance checks, data attached"
        # create the subject if not available
        if subject == None:
            content = f"[{self.args['run_date']}] Ledger Balance Alert"

        # write the result to tmp
        try:
            msg = ""
            tmpdir = tempfile.mkdtemp()
            fname = os.path.join(tmpdir, f'LEDGER-BALANCE-CHECK_{name}.csv')
            data.to_csv(fname, index=False)

            if send_attachment:
                attachments = [fname]
            else:
                attachments = []

            # send the email
            try:
                notify_success = True
                msg = f"Spec: {name}" + "\n"
                msg += f"Retries: {retries}\n"
                msg += f"Sender: {sender}" + "\n"
                msg += f"Receivers: {receivers}" + "\n"
                msg += f"Reply-To: {reply_to}" + "\n"
                msg += f"CC: {cc}" + "\n"
                msg += f"BCC: {bcc}" + "\n"
                msg += f"Subject: {subject}" + "\n"
                msg += f"Content: {content}" + "\n"

                send_html_email(content=content,
                                sender=sender,
                                receivers=receivers,
                                cc=cc,
                                bcc=bcc,
                                reply_to=reply_to,
                                subject=subject,
                                attachments=attachments,
                                cred=smtpcred,
                                retries=retries)
                logger.debug(
                    f"Sent notifications",
                    extra={"transform": self.name, "data": msg}
                )
            except:
                logger.exception(
                    f"Error sending email",
                    extra={"transform": self.name, "data": msg}
                )
        except:
            logger.exception(
                f"Error sending email",
                extra={"transform": self.name, "data": msg}
            )

        if os.path.exists(tmpdir):
            try:
                shutil.rmtree(tmpdir)
            except:
                logger.exception(
                    f"Unable to cleanup tmpdir",
                    extra={"transform": self.name, 'data': tmpdir}
                )

        return notify_success


    def notify_target_common(self, notify_obj, spec):
        # common notification handler
        notify_channel = notify_obj['channel']
        notify_success = False

        if notify_channel == 'email':
            notify_success = self.notify_via_email(notify_obj, spec)
        elif notify_channel == 'slack':
            # send the slack notification
            notify_success = self.notify_via_slack(notify_obj, spec)
        else:
            logger.exception(
                f"Unsupported notification type, skipping",
                extra={"transform": self.name, "data": json.dumps(notify_obj, indent=4)}
            )

        return notify_success


    def notify_target(self, spec, target, results, data):
        config      = spec['config']
        store       = config['store']
        params      = target.get('params', {})

        status = {
            "stats": {
                "tries": 0,
                "success": 0,
            }
        }

        # check to see if a custom notification builder is available
        if target.get('method', 'default') != 'default':
            # check to see if custom handler is available
            handler = target.get('handler')
            if handler is None or not hasattr(self, handler):
                logger.exception(
                    f"No notification handler, skipping.",
                    extra={"transform": self.name, "data": json.dumps(target, indent=4)}
                )
                status['stats']['tries'] += 1
                return status
            # we have the notification callback
            callback = getattr(self, handler)

            # call the handler to set up the notification
            notify_set = callback(spec, target, results, data, params)
        else:
            notify_set = {
                "sent": False,
                "name": target['name'],
                "channel": target['channel'],
                "freq": target['freq'],
                "send_on": target['send_on'],
                "sets": [target]
            }

        status['channel'] = notify_set['channel']
        status['do_notify'] = {}

        # for each notify object in the set
        for notify_obj in notify_set['sets']:
            # check to see if we need to send notifications
            set_id = notify_obj.get('set_id', '__ALL__')
            do_notify = self.check_notify_default(notify_set, notify_obj, spec, results)
            status['do_notify'][set_id] = do_notify

            if do_notify and not notify_set['sent']:
                notify_obj['channel'] = notify_set['channel']
                status['stats']['tries'] += 1
                # we need to now send a notification, it's not been handled in the custom handler
                success = self.notify_target_common(notify_obj, spec)
                if success:
                    status['stats']['success'] += 1
            else:
                logger.debug(
                    f"[{set_id}] Notification not needed",
                    extra={"transform": self.name}
                )

        return status


    def notify_result(self, spec, results, data):
        # send notification
        name = spec['name']
        config = spec['config']
        checks = config['checks']
        store  = config['store']
        notify = config.get('notify')

        # validate notification spec
        if notify is None or not isinstance(notify, list) or len(notify) == 0:
            logger.debug(
                f"Skipping notification -- no targets specified",
                extra={"transform": self.name}
            )
            return results

        # validate store spec, we need it also
        for f in ["sink", "params"]:
            if f not in store:
                logger.exception(
                    f"Store has no {f} param, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                )
                return results

        # all good, we can proceed
        targets_notified = []
        results_expanded = []
        notified_at = datetime.now()

        # for each notification target
        for target in notify:

            skip_target = False

            for f in ["method"]:
                if f not in target:
                    logger.exception(
                        f"Target has no {f} param, skipping.",
                        extra={"transform": self.name, "data": json.dumps(target, indent=4)}
                    )
                    skip_target = True
            if skip_target:
                return results

            # process the notification for each check in the spec
            for check in checks:

                r = results[results['check']==check['name']].copy()
                notify_status = self.notify_target(spec, target, r, data)

                # note it
                targets_notified.append({
                    check['name']: notify_status
                })

                for set_id, is_notified in notify_status['do_notify'].items():
                    # we need to make one copy per set_id
                    _r = r.copy()

                    # add this check+set to a list
                    _r['set_id']   = set_id
                    _r['notified'] = is_notified
                    _r['notifier'] = target['name']
                    _r['notified_at'] = None

                    if is_notified:
                        _r['notified_at'] = notified_at

                    results_expanded.append(_r)

        if len(targets_notified)>0:
            msg = f"Total {len(targets_notified)} checks notified: {targets_notified}" + "\n"
            logger.debug(
                f"Notifications processed",
                extra={"transform": self.name, "data": msg}
            )

        # concat all the sets of notification events for the results
        # we should have one per notification target
        results = pd.concat(results_expanded)

        return results


    def process(self, state):
        """
        Run the computation and update the state
        """
        logger.debug(
            "Start execution", extra=self.config.get_extra({"transform": self.name})
        )

        # Will be used in other places..
        self.state = state

        # Get the profile spec
        is_valid, profile, msg = profilespec.get_profile(self, "observability")
        if is_valid:
            logger.debug(
                f"Loaded profilespec",
                extra={"transform": self.name, "data": msg}
            )
        else:
            logger.error(
                f"Could not load profilespec",
                extra={"transform": self.name, "data": msg}
            )
            raise Exception("could not load profilespec")

        specs = profile.get("specs", None)
        if specs is None:
            raise Exception("Could not find 'specs' in profile")

        # get the dataset lookup table
        customer_datasets = profilespec.construct_dataset_list(self, specs)

        # Now go through each spec and process it
        for spec in specs:

            ## first, some checks on the spec
            do_process_spec = True
            name = spec.get('name', 'NO_SPEC_NAME')

            enabled = spec.get("enable", True)
            if not enabled:
                logger.debug(
                    f"Spec not enabled, skipping.",
                    extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                )
                do_process_spec = False
                continue

            for f in ["name", "config"]:
                if f not in spec:
                    logger.exception(
                        f"Spec has no {f} param, skipping.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                    )
                    do_process_spec = False
                    break
            if do_process_spec == False:
                continue

            config = spec['config']

            for f in ["source", "checks", "store"]:
                if f not in config:
                    logger.exception(
                        f"Spec config has no {f} param, skipping.",
                        extra={"transform": self.name, "data": json.dumps(spec, indent=4)}
                    )
                    do_process_spec = False
                    break
            if do_process_spec == False:
                continue


            ## we can now proceed with processing the spec
            # frist, load the source data
            data = self.load_dataset(spec, customer_datasets)

            # then, process it
            results = self.process_spec(spec, data)
            if results is None:
                continue

            ## notify the observability result
            results = self.notify_result(spec, results, data)

            ## store the observability result and notification status
            self.store_result(spec, results, data)

            # update frame for pipline
            description = spec.get("desc", f"{name} observability results")
            lineage = {
                "type": "lineage",
                "transform": self.name,
                "dependencies": [
                    {
                        "type": "file",
                        "nature": "input",
                        "objects": [spec.get("filename", "__NEW__")],
                    },
                ],
            }
            self.update_frame(
                spec,
                description,
                results,
                lineage,
            )

        # Done
        logger.debug(
            "Complete execution", extra=self.config.get_extra({"transform": self.name})
        )

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        pass



import json
from collections import OrderedDict

from django.db import models
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from jsonfield import JSONField
from django.utils.translation import gettext_lazy as _

from enrichsdk.utils import SafeEncoder

class AppPolicyBase(models.Model):
    """
    Common policy document
    """

    class Meta:
        abstract = True

    appname    = models.CharField(max_length=64,
                                  default="default",
                                  verbose_name="App Name")

    namespace    = models.CharField(max_length=64,
                                    default="default",
                                    verbose_name="Namespace (to support multiple instances)")


    name         = models.CharField(max_length=256,
                                    blank=False,
                                    verbose_name="Name")
    desc         = models.CharField(max_length=1024,
                                    blank=False,
                                    verbose_name="Description")
    notes        = models.CharField(max_length=1024,
                                    blank=True,
                                    verbose_name="Additional Notes",
                                    help_text="")
    tags         = models.CharField(max_length=128,
                                    blank=True,
                                    default="",
                                    verbose_name="Tags")
    active       = models.BooleanField(default=True,
                                  verbose_name="Active",
                                  help_text="Enable/Disable")

    config = JSONField(verbose_name="Configuration", default={})
    # Hygiene
    created_at   = models.DateTimeField(auto_now_add=True)
    created_by   = models.ForeignKey(User, related_name='%(app_label)s_%(class)s_creator', on_delete=models.PROTECT)
    modified_by  = models.ForeignKey(User, related_name='%(app_label)s_%(class)s_modifier', on_delete=models.PROTECT)
    modified_at  = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.desc

    def export(self, spec):

        response = OrderedDict([
            ('schema', self.__class__.__name__),
            ('id', self.pk),
            ('app', self.appname),
            ('name', self.name),
            ('namespace', self.namespace),
            ('desc', self.desc),
            ('notes', self.notes),
            ('tags', self.tags),
            ('active', self.active),
            ('created_at', self.created_at.isoformat()),
            ('created_by', self.created_by.username),
            ('modified_at', self.modified_at.isoformat()),
            ('modified_by', self.modified_by.username),
            ('config', self.config)
        ])

        return response


